from nyxstone import Nyxstone, Instruction, IntegerBase

nyxstone = Nyxstone("x86_64")

# basic functionality
assert nyxstone.assemble("mov rax, rax") == [0x48, 0x89, 0xC0]
assert nyxstone.disassemble([0x48, 0x89, 0xC0]) == "mov rax, rax\n"
# reordering of arguments
assert nyxstone.assemble(
    address=0x1000, labels={".label": 0x1200}, assembly="mov rax, rax"
) == [0x48, 0x89, 0xC0]

# inline/external labels
assert nyxstone.assemble("jmp .label; nop; .label:") == [0xEB, 0x01, 0x90]
assert nyxstone.assemble("jmp .label", labels={".label": 0x1000}) == [
    0xE9,
    0xFB,
    0x0F,
    0x00,
    0x00,
]
assert nyxstone.assemble_to_instructions("jmp .label", 0x1000, {".label": 0x1200}) == [
    Instruction(0x1000, "jmp .label", [0xE9, 0xFB, 0x01, 0x00, 0x00])
]

# disassembling using address and count
assert nyxstone.disassemble_to_instructions(
    [0x48, 0x31, 0xC0, 0x48, 0x01, 0xD8], 0x1000, 0
) == [
    Instruction(0x1000, "xor rax, rax", [0x48, 0x31, 0xC0]),
    Instruction(0x1003, "add rax, rbx", [0x48, 0x01, 0xD8]),
]
assert nyxstone.disassemble_to_instructions(
    [0x48, 0x31, 0xC0, 0x48, 0x01, 0xD8], 0x1000, 1
) == [Instruction(0x1000, "xor rax, rax", [0x48, 0x31, 0xC0])]

# specify additional features
nyxstone = Nyxstone("thumbv8", features="+mve.fp,+fp16")

nyxstone.assemble("vadd.f16 s0, s1, s2")

# specify the immediate style
nyxstone = Nyxstone("thumbv8", immediate_style=IntegerBase.HexPrefix)
assert nyxstone.assemble_to_instructions("add r0, r0, #1") == [
    Instruction(0x0, "add.w r0, r0, #0x1", [0x00, 0xF1, 0x01, 0x00])
]
nyxstone = Nyxstone("thumbv8", immediate_style=IntegerBase.HexSuffix)
assert nyxstone.assemble_to_instructions("add r0, r0, #1") == [
    Instruction(0x0, "add.w r0, r0, #1h", [0x00, 0xF1, 0x01, 0x00])
]

# handling an error:
try:
    invalid = nyxstone.assemble("mov r20, r20")
    assert false, "Unreachable"
except Exception as e:
    assert (
        str(e)
        == "Error during assembly: error: operand must be a register in range [r0, r14]\nmov r20, r20\n    ^\n"
    )



import nyxstone_cpp
from enum import IntEnum


class IntegerBase(IntEnum):
    """Styles for immediate printing in Nyxstone.

    Members
    -------
    Dec:
        Decimal printing
    HexPrefix:
        Hex, prefixed with '0x'
    HexSuffix:
        Hex, suffixed with 'h'
    """

    Dec = nyxstone_cpp.IntegerBase.Dec
    HexPrefix = nyxstone_cpp.IntegerBase.HexPrefix
    HexSuffix = nyxstone_cpp.IntegerBase.HexSuffix


class Instruction(nyxstone_cpp.Instruction):
    """
    Extended information of an Instruction.
    """

    @classmethod
    def _from_cpp_instruction(cls, instr: nyxstone_cpp.Instruction):
        return cls(instr.address, instr.assembly, instr.bytes)


class Nyxstone:
    """
    Nyxstone class used for assembling and disassembling for a given architecture.
    """

    def __init__(
        self,
        target_triple: str,
        cpu: str = "",
        features: str = "",
        immediate_style: IntegerBase = IntegerBase.Dec,
    ):
        """
        Parameters
        ----------
        target_triple : str
            Llvm target triple or architecture identifier of triple.
        cpu : str, optional
            Llvm CPU specifier (defaults to no specific CPU).
        features : str, optional
            Llvm feature string. The feature string is a comma seperated list of
            features which are prepended with a '+' for enabling and a '-' for disabling (defaults to no features).
        immediate_style : IntegerBase, optional
            Printing style of immediates in disassembly and extended instruction details.

        Raises
        ------
        ValueError
            If the creation of the nyxstone instance failed.
        """

        res = nyxstone_cpp.create_nyxstone(
            target_triple, cpu, features, nyxstone_cpp.IntegerBase(immediate_style)
        )
        if isinstance(res, nyxstone_cpp.NyxstoneError):
            raise ValueError(res.err)
        self.nyxstone = res

    def assemble(
        self, assembly: str, address: int = 0x0, labels: dict[str, int] = {}
    ) -> list[int]:
        """Translates assembly instructions at a given start address to bytes.

        Additional label definitions by absolute address may be supplied.
        Does not support assembly directives that impact the layout (f. i., .section, .org).

        Parameters
        ----------
        assembly : str
            The assembly instruction(s) to be assembled.
        address : int, optional
            The absolute address of the first instruction (default is 0x0).
        labels : dict[str, int], optional
            Label definitions, should hold all external labels used in assembly (default is an empty dict).

        Raises
        ------
        ValueError
            If assembling failed.

        Returns
        -------
        list[int]
            A list of 8-bit integers representing the assembled byte code.
        """

        res = self.nyxstone.assemble(assembly, address, labels)
        if isinstance(res, nyxstone_cpp.NyxstoneError):
            raise ValueError(res.err)
        return res

    def assemble_to_instructions(
        self, assembly: str, address: int = 0x0, labels: dict[str, int] = {}
    ) -> list[Instruction]:
        """Translates assembly instructions at given start address to instruction details containing bytes.

        Additional label definitions by absolute address may be supplied.
        Does not support assembly directives that impact the layout (f. i., .section, .org).

        Parameters
        ----------
        assembly : str
            The assembly instruction(s) to be assembled.
        address : int, optional
            The absolute address of the first instruction (default is 0x0).
        labels : dict[str, int], optional
            Label definitions, should hold all external labels used in assembly (default is an empty dict).

        Raises
        ------
        ValueError
            If assembling failed.

        Returns
        -------
        list[Instruction]
            A list of instruction details.
        """

        res = self.nyxstone.assemble_to_instructions(assembly, address, labels)
        if isinstance(res, nyxstone_cpp.NyxstoneError):
            raise ValueError(res.err)
        return list(map(Instruction._from_cpp_instruction, res))

    def disassemble(
        self, bytecode: list[int], address: int = 0x0, count: int = 0
    ) -> str:
        """Translates bytes to disassembly text at given start address.

        Parameters
        ----------
        bytes : list[int]
            The byte code to be disassembled.
        address : int, optional
            The absolute address of the byte code (default is 0x0).
        count : int, optional
            The number of instructions which should be disassembled, 0 means all (default is 0).

        Raises
        ------
        ValueError
            If disassembling failed.

        Returns
        -------
        str
            The disassembled text.
        """

        res = self.nyxstone.disassemble(bytecode, address, count)
        if isinstance(res, nyxstone_cpp.NyxstoneError):
            raise ValueError(res.err)
        return res

    def disassemble_to_instructions(
        self, bytecode: list[int], address: int = 0x0, count: int = 0
    ) -> list[Instruction]:
        """Translates bytes to instruction details containing disassembly text at given start address.

        Parameters
        ----------
        bytes : list[int]
            The byte code to be disassembled.
        address : int, optional
            The absolute address of the byte code (default is 0x0).
        count : int, optional
            The number of instructions which should be disassembled, 0 means all (default is 0).

        Raises
        ------
        ValueError
            If disassembling failed.

        Returns
        -------
        list[Instruction]
            A list of instruction details.
        """

        res = self.nyxstone.disassemble_to_instructions(bytecode, address, count)
        if isinstance(res, nyxstone_cpp.NyxstoneError):
            raise ValueError(res.err)
        return list(map(Instruction._from_cpp_instruction, res))



from glob import glob
from setuptools import setup
import os
import subprocess

try:
    from pybind11.setup_helpers import Pybind11Extension
except ImportError:
    from setuptools import Extension as Pybind11Extension

LLVM_PREFIX_NAME = "NYXSTONE_LLVM_PREFIX"


class ValidLLVMConfig:
    def __init__(self):
        llvm_config = (
            os.path.join(os.getenv(LLVM_PREFIX_NAME), "bin/llvm-config")
            if os.getenv(LLVM_PREFIX_NAME) != None
            else "llvm-config"
        )

        try:
            version = subprocess.run(
                [llvm_config, "--version"], capture_output=True
            ).stdout.decode("utf-8")
        except FileNotFoundError as e:
            print(f"Could not find llvm-config in ${LLVM_PREFIX_NAME} or $PATH")
            exit(1)

        major_version = version.split(".")[0]

        if major_version != "15":
            print(
                f"LLVM Major version must be 15, found {major_version}! Try setting the env variable ${LLVM_PREFIX_NAME} to tell nyxstone about the install location of LLVM 15."
            )
            exit(1)

        lib_output = subprocess.run(
            [
                llvm_config,
                "--link-static",
                "--libs",
            ],
            capture_output=True,
        )

        if lib_output.returncode != 0:
            print(
                "Cannot link statically, please install LLVM with static linking support"
            )
            exit(1)

        self.config = llvm_config

    def get_llvm_libraries(self) -> (list[str], str):
        lib_output = subprocess.run(
            [
                self.config,
                "--link-static",
                "--system-libs",
                "--libs",
                "core",
                "mc",
                "all-targets",
            ],
            capture_output=True,
        )

        libraries = (
            lib_output.stdout.decode("utf-8")
            .replace("\n", " ")
            .strip()
            .replace("-l", "")
            .split(" ")
        )

        lib_location = (
            subprocess.run([self.config, "--ldflags"], capture_output=True)
            .stdout.decode("utf-8")
            .replace("\n", "")  # remove trailing '\n'
            .replace("-L", "")  # remove leading -L re-added by pybind
            .strip()  # strip trailing whitespace
        )

        return libraries, lib_location

    def get_llvm_include_dir(self) -> str:
        llvm_include_dir = (
            subprocess.run([self.config, "--includedir"], capture_output=True)
            .stdout.decode("utf-8")
            .replace("\n", "")
            .strip()
        )

        return llvm_include_dir


llvm = ValidLLVMConfig()
llvm_libs, llvm_lib_dir = llvm.get_llvm_libraries()
llvm_inc_dir = llvm.get_llvm_include_dir()

nyxstone_srcs = glob("nyxstone-cpp/src/*.cpp")

binder = ["nyxstone-cpp/pynyxstone.cpp"]

srcs = sorted(nyxstone_srcs + binder)

desc = "Python bindings for the nyxstone library."

ext_modules = [
    Pybind11Extension(
        name="nyxstone_cpp",
        sources=srcs,
        include_dirs=["nyxstone-cpp/include/", "nyxstone-cpp/src/", llvm_inc_dir],
        libraries=llvm_libs,
        library_dirs=[llvm_lib_dir],
        extra_link_args=[
            "-static-libstdc++",
            "-static-libgcc",
        ],  # try to reduce dependencies
    )
]

setup(
    name="nyxstone",
    version="0.1.0",
    description=desc,
    author="emproof B.V.",
    author_email="oss@emproof.com",
    packages=["nyxstone"],
    ext_modules=ext_modules,
)
